"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var fs_1 = require("fs");
var require_relative_1 = __importDefault(require("require-relative"));
var mergeOptions_1 = __importDefault(require("../../../src/utils/mergeOptions"));
var relativeId_1 = require("../../../src/utils/relativeId");
var logging_1 = require("../logging");
var batchWarnings_1 = __importDefault(require("./batchWarnings"));
var build_1 = __importDefault(require("./build"));
var loadConfigFile_1 = __importDefault(require("./loadConfigFile"));
var watch_1 = __importDefault(require("./watch"));
function runRollup(command) {
    var inputSource;
    if (command._.length > 0) {
        if (command.input) {
            logging_1.handleError({
                code: 'DUPLICATE_IMPORT_OPTIONS',
                message: 'use --input, or pass input path as argument'
            });
        }
        inputSource = command._;
    }
    else if (typeof command.input === 'string') {
        inputSource = [command.input];
    }
    else {
        inputSource = command.input;
    }
    if (inputSource && inputSource.length > 0) {
        if (inputSource.some(function (input) { return input.indexOf('=') !== -1; })) {
            command.input = {};
            inputSource.forEach(function (input) {
                var equalsIndex = input.indexOf('=');
                var value = input.substr(equalsIndex + 1);
                var key = input.substr(0, equalsIndex);
                if (!key)
                    key = relativeId_1.getAliasName(input);
                command.input[key] = value;
            });
        }
        else {
            command.input = inputSource;
        }
    }
    if (command.environment) {
        var environment = Array.isArray(command.environment)
            ? command.environment
            : [command.environment];
        environment.forEach(function (arg) {
            arg.split(',').forEach(function (pair) {
                var _a = pair.split(':'), key = _a[0], value = _a[1];
                if (value) {
                    process.env[key] = value;
                }
                else {
                    process.env[key] = String(true);
                }
            });
        });
    }
    var configFile = command.config === true ? 'rollup.config.js' : command.config;
    if (configFile) {
        if (configFile.slice(0, 5) === 'node:') {
            var pkgName = configFile.slice(5);
            try {
                configFile = require_relative_1.default.resolve("rollup-config-" + pkgName, process.cwd());
            }
            catch (err) {
                try {
                    configFile = require_relative_1.default.resolve(pkgName, process.cwd());
                }
                catch (err) {
                    if (err.code === 'MODULE_NOT_FOUND') {
                        logging_1.handleError({
                            code: 'MISSING_EXTERNAL_CONFIG',
                            message: "Could not resolve config file " + configFile
                        });
                    }
                    throw err;
                }
            }
        }
        else {
            // find real path of config so it matches what Node provides to callbacks in require.extensions
            configFile = fs_1.realpathSync(configFile);
        }
        if (command.watch)
            process.env.ROLLUP_WATCH = 'true';
        loadConfigFile_1.default(configFile, command)
            .then(function (configs) { return execute(configFile, configs, command); })
            .catch(logging_1.handleError);
    }
    else {
        return execute(configFile, [{ input: null }], command);
    }
}
exports.default = runRollup;
function execute(configFile, configs, command) {
    if (command.watch) {
        watch_1.default(configFile, configs, command, command.silent);
    }
    else {
        var promise = Promise.resolve();
        var _loop_1 = function (config) {
            promise = promise.then(function () {
                var warnings = batchWarnings_1.default();
                var _a = mergeOptions_1.default({
                    command: command,
                    config: config,
                    defaultOnWarnHandler: warnings.add
                }), inputOptions = _a.inputOptions, outputOptions = _a.outputOptions, optionError = _a.optionError;
                if (optionError)
                    inputOptions.onwarn({ code: 'UNKNOWN_OPTION', message: optionError });
                return build_1.default(inputOptions, outputOptions, warnings, command.silent);
            });
        };
        for (var _i = 0, configs_1 = configs; _i < configs_1.length; _i++) {
            var config = configs_1[_i];
            _loop_1(config);
        }
        return promise;
    }
}
