"use strict";
var __spreadArrays = (this && this.__spreadArrays) || function () {
    for (var s = 0, i = 0, il = arguments.length; i < il; i++) s += arguments[i].length;
    for (var r = Array(s), k = 0, i = 0; i < il; i++)
        for (var a = arguments[i], j = 0, jl = a.length; j < jl; j++, k++)
            r[k] = a[j];
    return r;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var ExportDefaultVariable_1 = __importDefault(require("../ast/variables/ExportDefaultVariable"));
var ExternalModule_1 = __importDefault(require("../ExternalModule"));
var reservedNames_1 = require("./reservedNames");
var safeName_1 = require("./safeName");
var DECONFLICT_IMPORTED_VARIABLES_BY_FORMAT = {
    amd: deconflictImportsOther,
    cjs: deconflictImportsOther,
    es: deconflictImportsEsm,
    iife: deconflictImportsOther,
    system: deconflictImportsEsm,
    umd: deconflictImportsOther
};
function deconflictChunk(modules, dependencies, imports, usedNames, format, interop, preserveModules) {
    var _a = reservedNames_1.RESERVED_NAMES_BY_FORMAT[format], forbiddenNames = _a.forbiddenNames, formatGlobals = _a.formatGlobals;
    Object.assign(usedNames, forbiddenNames);
    Object.assign(usedNames, formatGlobals);
    addUsedGlobalNames(usedNames, modules);
    deconflictTopLevelVariables(usedNames, modules);
    DECONFLICT_IMPORTED_VARIABLES_BY_FORMAT[format](usedNames, imports, dependencies, interop, preserveModules);
    for (var _i = 0, modules_1 = modules; _i < modules_1.length; _i++) {
        var module_1 = modules_1[_i];
        module_1.scope.deconflict(forbiddenNames);
    }
}
exports.deconflictChunk = deconflictChunk;
function addUsedGlobalNames(usedNames, modules) {
    var accessedGlobals = Object.assign.apply(Object, __spreadArrays([{}], modules.map(function (module) { return module.scope.accessedOutsideVariables; })));
    for (var _i = 0, _a = Object.keys(accessedGlobals); _i < _a.length; _i++) {
        var name_1 = _a[_i];
        var variable = accessedGlobals[name_1];
        if (variable.included) {
            usedNames[name_1] = true;
        }
    }
}
function deconflictImportsEsm(usedNames, imports, _dependencies, interop) {
    for (var _i = 0, _a = Array.from(imports); _i < _a.length; _i++) {
        var variable = _a[_i];
        var module_2 = variable.module;
        var name_2 = variable.name;
        var proposedName = void 0;
        if (module_2 instanceof ExternalModule_1.default && (name_2 === '*' || name_2 === 'default')) {
            if (name_2 === 'default' && interop && module_2.exportsNamespace) {
                proposedName = module_2.variableName + '__default';
            }
            else {
                proposedName = module_2.variableName;
            }
        }
        else {
            proposedName = name_2;
        }
        variable.setRenderNames(null, safeName_1.getSafeName(proposedName, usedNames));
    }
}
function deconflictImportsOther(usedNames, imports, dependencies, interop, preserveModules) {
    for (var _i = 0, dependencies_1 = dependencies; _i < dependencies_1.length; _i++) {
        var chunkOrExternalModule = dependencies_1[_i];
        chunkOrExternalModule.variableName = safeName_1.getSafeName(chunkOrExternalModule.variableName, usedNames);
    }
    for (var _a = 0, _b = Array.from(imports); _a < _b.length; _a++) {
        var variable = _b[_a];
        var module_3 = variable.module;
        if (module_3 instanceof ExternalModule_1.default) {
            var name_3 = variable.name;
            if (name_3 === 'default' && interop && (module_3.exportsNamespace || module_3.exportsNames)) {
                variable.setRenderNames(null, module_3.variableName + '__default');
            }
            else if (name_3 === '*' || name_3 === 'default') {
                variable.setRenderNames(null, module_3.variableName);
            }
            else {
                variable.setRenderNames(module_3.variableName, null);
            }
        }
        else {
            var chunk = module_3.chunk;
            if (chunk.exportMode === 'default' || (preserveModules && variable.isNamespace)) {
                variable.setRenderNames(null, chunk.variableName);
            }
            else {
                variable.setRenderNames(chunk.variableName, module_3.chunk.getVariableExportName(variable));
            }
        }
    }
}
function deconflictTopLevelVariables(usedNames, modules) {
    for (var _i = 0, modules_2 = modules; _i < modules_2.length; _i++) {
        var module_4 = modules_2[_i];
        var moduleVariables = module_4.scope.variables;
        for (var _a = 0, _b = Object.keys(moduleVariables); _a < _b.length; _a++) {
            var name_4 = _b[_a];
            var variable = moduleVariables[name_4];
            if (variable.included &&
                // this will only happen for exports in some formats
                !(variable.renderBaseName ||
                    (variable instanceof ExportDefaultVariable_1.default && variable.referencesOriginal()))) {
                variable.setRenderNames(null, safeName_1.getSafeName(variable.name, usedNames));
            }
        }
        var namespace = module_4.getOrCreateNamespace();
        if (namespace.included) {
            namespace.setRenderNames(null, safeName_1.getSafeName(namespace.name, usedNames));
        }
    }
}
